/**
 *    Copyright (C) 2016 MongoDB Inc.
 *
 *    This program is free software: you can redistribute it and/or  modify
 *    it under the terms of the GNU Affero General Public License, version 3,
 *    as published by the Free Software Foundation.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Affero General Public License for more details.
 *
 *    You should have received a copy of the GNU Affero General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    As a special exception, the copyright holders give permission to link the
 *    code of portions of this program with the OpenSSL library under certain
 *    conditions as described in each individual source file and distribute
 *    linked combinations including the program with the OpenSSL library. You
 *    must comply with the GNU Affero General Public License in all respects
 *    for all of the code used other than as permitted herein. If you modify
 *    file(s) with this exception, you may extend this exception to your
 *    version of the file(s), but you are not obligated to do so. If you do not
 *    wish to do so, delete this exception statement from your version. If you
 *    delete this exception statement from all source files in the program,
 *    then also delete it in the license file.
 */

#include "mongo/platform/basic.h"

#include "mongo/bson/bsonmisc.h"
#include "mongo/bson/bsonobjbuilder.h"
#include "mongo/s/request_types/commit_chunk_migration_request_type.h"
#include "mongo/unittest/unittest.h"

namespace mongo {

using unittest::assertGet;

namespace {

const auto kNamespaceString = NamespaceString("TestDB", "TestColl");
const auto kShardHasDistributedLock = false;

const auto kShardId0 = ShardId("shard0");
const auto kShardId1 = ShardId("shard1");

const auto kKey0 = BSON("Key" << -100);
const auto kKey1 = BSON("Key" << 100);
const auto kKey2 = BSON("Key" << -50);
const auto kKey3 = BSON("Key" << 50);

const char kConfigSvrCommitChunkMigration[] = "_configsvrCommitChunkMigration";

TEST(CommitChunkMigrationRequest, WithControlChunk) {
    BSONObjBuilder builder;

    ChunkType migratedChunkType;
    migratedChunkType.setMin(kKey0);
    migratedChunkType.setMax(kKey1);

    ChunkType controlChunkTypeTemp;
    controlChunkTypeTemp.setMin(kKey2);
    controlChunkTypeTemp.setMax(kKey3);
    boost::optional<ChunkType> controlChunkType = std::move(controlChunkTypeTemp);

    ChunkVersion fromShardCollectionVersion(1, 2, OID::gen());

    CommitChunkMigrationRequest::appendAsCommand(&builder,
                                                 kNamespaceString,
                                                 kShardId0,
                                                 kShardId1,
                                                 migratedChunkType,
                                                 controlChunkType,
                                                 fromShardCollectionVersion,
                                                 kShardHasDistributedLock);

    BSONObj cmdObj = builder.obj();

    auto request = assertGet(CommitChunkMigrationRequest::createFromCommand(
        NamespaceString(cmdObj[kConfigSvrCommitChunkMigration].String()), cmdObj));

    ASSERT_EQ(kNamespaceString, request.getNss());
    ASSERT_EQ(kShardId0, request.getFromShard());
    ASSERT_EQ(kShardId1, request.getToShard());
    ASSERT_BSONOBJ_EQ(kKey0, request.getMigratedChunkRange().getMin());
    ASSERT_BSONOBJ_EQ(kKey1, request.getMigratedChunkRange().getMax());
    ASSERT(request.hasControlChunkRange());
    ASSERT_BSONOBJ_EQ(kKey2, request.getControlChunkRange().getMin());
    ASSERT_BSONOBJ_EQ(kKey3, request.getControlChunkRange().getMax());
    ASSERT_EQ(fromShardCollectionVersion, request.getFromShardCollectionVersion());
    ASSERT_EQ(kShardHasDistributedLock, request.shardHasDistributedLock());
}

TEST(CommitChunkMigrationRequest, WithoutControlChunk) {
    BSONObjBuilder builder;

    ChunkType migratedChunkType;
    migratedChunkType.setMin(kKey0);
    migratedChunkType.setMax(kKey1);

    ChunkVersion fromShardCollectionVersion(1, 2, OID::gen());

    CommitChunkMigrationRequest::appendAsCommand(&builder,
                                                 kNamespaceString,
                                                 kShardId0,
                                                 kShardId1,
                                                 migratedChunkType,
                                                 boost::none,
                                                 fromShardCollectionVersion,
                                                 kShardHasDistributedLock);

    BSONObj cmdObj = builder.obj();

    auto request = assertGet(CommitChunkMigrationRequest::createFromCommand(
        NamespaceString(cmdObj[kConfigSvrCommitChunkMigration].String()), cmdObj));

    ASSERT_EQ(kNamespaceString, request.getNss());
    ASSERT_EQ(kShardId0, request.getFromShard());
    ASSERT_EQ(kShardId1, request.getToShard());
    ASSERT_BSONOBJ_EQ(kKey0, request.getMigratedChunkRange().getMin());
    ASSERT_BSONOBJ_EQ(kKey1, request.getMigratedChunkRange().getMax());
    ASSERT(!request.hasControlChunkRange());
    ASSERT_EQ(fromShardCollectionVersion, request.getFromShardCollectionVersion());
    ASSERT_EQ(kShardHasDistributedLock, request.shardHasDistributedLock());
}

}  // namespace
}  // namespace mongo
